#----------------------------------------------------------------
# Copyright (c) 2002 Benjamin Crowell, all rights reserved.
#
# This software is available under version 2 of the GPL license.
# The software is copyrighted, and you must agree to the
# license in order to have permission to copy it. The full
# text of the license is given in the file titled Copying.
#
#----------------------------------------------------------------


use strict;

use GradeBook;
use Crunch;
use Text;
use NetOG;

package Report;

sub stats {
    my %args = (
		GB=>"",
		FORMAT=>"plain",
                @_,
                );
    my $gb = $args{GB};
    my $format = $args{FORMAT};
    my $t = Text->new($format);

    my $h = Crunch::class_stats($gb,"all");
    my $possible = Crunch::possible($gb,"all");
    my %h = %$h;
    foreach my $c(keys(%h)) {
      my $stats = $h{$c};
      my ($n,$mean,$sd) = (
        GradeBook::get_property($stats,"n"),
        GradeBook::get_property($stats,"mean"),
        GradeBook::get_property($stats,"sd"),
      );
      #print "$c possible = ".$possible->{$c}."\n";
      my $p = $possible->{$c};
      my $stuff = "";
      if ($p > 0.) {
        $stuff = (sprintf "n=%3d   mean=%7.2f%s   sd=%7.2f%s",
                        $n,$mean/$p*100.0,"%",$sd/$p*100.0,"%"
                  );
      }
      else {
        $stuff = "";
      }
      $t->put(TEXT=>(sprintf "%10s   %s\n",$c,$stuff));
    }

    return $t;
}


sub sort {
    my $gb = shift;
    my $format = shift;
    my $cat = '';
    if (@_) {$cat=shift}
    my $ass = '';
    if (@_) {$ass=shift}
    my $sort_by = 'score';
    if ($format=~m/\*$/) {$sort_by='name'; $format=~s/\*$//} # kludge: trailing * means sort by name
    my @student_keys = $gb->student_keys();
    my $t = Text->new($format);
    my @stuff = ();
    my $title;
    if ($cat eq '') {$title=''}
    if ($cat ne '' && $ass eq '') {$title=$gb->category_name_plural($cat)}
    if ($ass ne '') {$title=$gb->category_name_singular($cat)." ".$ass}
    foreach my $student (@student_keys) {
      my $score;
      if ($cat eq '') {$score = Crunch::totals($gb,$student)->{'all'}; $title=''}
      if ($cat ne '' && $ass eq '') {$score = Crunch::totals($gb,$student,$cat)->{$cat}}
      if ($ass ne '') {$score = $gb->get_current_grade($student,$cat,$ass)."/".$gb->assignment_property("$cat.$ass","max")}
      $score =~ m@^([^/]*)@.*@;
      my $points = $1;
      push @stuff,[$student,$points,$score];
    }
    if ($sort_by eq 'score') {
      @stuff = sort {$b->[1] <=> $a->[1]} @stuff;
    }
    else { # sort by name
      @stuff = sort {$gb->compare_names($b->[0],$a->[0])} @stuff;
    }
    $t->put(P=>1,TEXT=>$title);
    foreach my $thing (@stuff) {
      my $student = $thing->[0];
      my $score = $thing->[2];
      my ($first,$last) = $gb->name($student);
      $t->put(BR=>1,TEXT=>(sprintf "%-20s %10s   (%s)" , "$first $last", fraction_to_display($gb,$score),format_fraction($score)));
    }
    return $t;
}

# Returns a Text object
sub class_totals {
    my $gb = shift;
    my $format = shift;
    my @student_keys = $gb->student_keys();
    my $t = Text->new($format);
    foreach my $student (@student_keys) {
	my ($first,$last) = $gb->name($student);
	my $full_name = $first." ".$last;
	my $totals = Crunch::totals($gb,$student);
        my $all = $totals->{"all"};
	$t->put(TEXT=>(sprintf "%20s %s\n",$full_name,
		       fraction_to_display($gb,$all)." ($all)"));
    }
    $t->put(BR=>1,TEXT=>(sprintf "%20s %d students","",1+$#student_keys));
    return $t;
}


sub make_student_report {
  my $gb = shift;
  my $who = shift;
  my ($first,$last) = $gb->name($who);
  my $t = student(GB=>$gb,STUDENT=>$who,FORMAT=>"html");
  return "<html><head><title>Grades for $first $last</title></head><body>\n"
               . $t->text()
               . "</body></html>\n";
}

=head3 upload_grades()

	GB=>'',
	PWD=>'',
        PROGRESS_BAR_CALLBACK=>sub{},
        FINAL_CALLBACK=>sub{},
        PROTOCOL=>'og',

This is meant to be something we can call from either the GUI or the text-based
interface, without duplicating any code. No other protocols are supported besides the default one.

=cut

sub upload_grades {
      my %args = (
	GB=>'',
	PWD=>'',
        PROGRESS_BAR_CALLBACK=>sub{},
        FINAL_CALLBACK=>sub{},
        PROTOCOL=>'og',
        @_
      );
      my $gb = $args{GB};
      my $password = $args{PWD};
      my $progress_bar = $args{PROGRESS_BAR_CALLBACK};
      my $protocol = $args{PROTOCOL};
      my $dir = $gb->dir(); # e.g. bcrowell/s2002/205 ; relative to cgi-bin
      my $final_callback = $args{FINAL_CALLBACK};
      my $result = '';
      #my $backup = $gb->boolean_pref("backups_on_server"); # <-- messed up, doesn't default to true
      my $backup = 1;
      if ($protocol eq 'og') {
        my $request = NetOG->new();
        my $prefs = $gb->preferences();
        my $server_domain = $prefs->get('server_domain');
        my $server_user = $prefs->get('server_user');
        my $server_account = $prefs->get('server_account');
        my $server_key = $prefs->get('server_key');
        my $server_class = $gb->dir();
        my $ndone = 0;
        my @student_keys = $gb->student_keys();
        foreach my $student(@student_keys) {
          $request->be_client(GB=>$gb,
            HOST=>$server_domain,SERVER_KEY=>$server_key,
            PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
                   'what'=>'upload_grade_report','who'=>$student,'report'=>make_student_report($gb,$student)}
          );
	  my $result = $request->{RESPONSE_DATA};
          ++$ndone;
          if (ref $progress_bar) {&$progress_bar($ndone/($#student_keys+1))}
	}
        if ($backup) {
          my $backup_file_name = $gb->file_name()."-".DateOG::current_date_sortable();
	  my $file_contents = '';
	  open(FILE,"<".($gb->file_name()));
	  while (my $line = <FILE>) {
	    $file_contents = $file_contents . $line;
	  }
	  close(FILE);
          $request->be_client(GB=>$gb,
            HOST=>$server_domain,SERVER_KEY=>$server_key,DATA=>$file_contents,
            PARS=>{'account'=>$server_account,'user'=>$server_user,'class'=>$server_class,
                   'what'=>'back_up_gradebook_file','name'=>$backup_file_name}
          );
	  #print $request->{RESPONSE_DATA}."\n";
        }
        &$final_callback($result);
      }
      return $result;
}

sub format_percent_letter_and_fraction {
        my $gb = shift;
        my $frac = shift;
        return fraction_to_display($gb,$frac)." (".format_fraction($frac).")";
}

sub format_fraction {
        my $frac = shift;
        if ($frac eq '' || $frac eq '--') {return $frac}
	my ($total,$possible) = split("/",$frac);
	$possible = sprintf "%d",int($possible);
	$total = sprintf "%".length(int($possible))."d",int($total);
	return sprintf "$total/$possible";
}

sub fraction_to_display {
        my $gb = shift;
        my $frac = shift;
	my ($percentage,$letter) = fraction_to_pct_and_letter($gb,$frac);
	if ($percentage eq "" && $letter eq "") {return ""}
	if ($letter eq "") {return sprintf "%4.1f",$percentage}
	if (length $letter<2) {$letter = "$letter "}
	return sprintf "%4.1f%s %s",$percentage,'%',$letter;
}


sub fraction_to_pct_and_letter {
        my $gb = shift;
        my $frac = shift;
        if ($frac eq '' || $frac eq '--') {return $frac}
	my ($total,$possible) = split("/",$frac);
	my $percentage = 0.0;
	my $letter = "";
	if ($possible>0) {
	    $percentage = 100.*$total/$possible;
	    $letter = $gb->percentage_to_letter_grade($percentage);
	}
	else {
	    $percentage = "";
	    $letter = "";
	}
	return ($percentage,$letter);
}

sub table {
    my %args = (
		GB=>"",
		FORMAT=>"plain",
                @_,
               );
    my $gb = $args{GB};
    my $format = $args{FORMAT};
    my @student_keys = $gb->student_keys();
    my $t = Text->new($format);

    my @a = split(",",$gb->assignment_list()); # list of cat.ass
    my @c = (); # cat only
    my @n = (); # ass only
    foreach my $a(@a) {
      my ($cat,$ass) = (GradeBook::first_part_of_label($a),
			      GradeBook::second_part_of_label($a));
      push @c,$cat;
      push @n,$ass;
    }
    my $n_assignments = @a;
    my $assignments_per_page = 12;
    my $width = 5;
    my $name_width = 15;
    my $full_width = $name_width+$assignments_per_page*$width;
    my $n_pages = int($n_assignments/$assignments_per_page);
    if ($n_pages * $assignments_per_page < $n_assignments) {$n_pages++}

    # Helper routine to trim or pad a string to a desired width:
    my $trim = sub {
      my $string = shift;
      my $width = shift;
      my $result = sprintf "%${width}s",$string;
      if (length $result>$width) {$result = substr($result,0,$width)}
      return $result;
    };

    for (my $page=0; $page<$n_pages; $page++) {
      my $offset = $page*$assignments_per_page;
      my $n_cols = $assignments_per_page;
      if ($offset+$n_cols>$n_assignments) {$n_cols = $n_assignments-$offset}

      $t->put(TEXT=>(' ' x $name_width));
      for (my $col=0; $col<$n_cols; $col++) {
	$t->put(TEXT=>&$trim($c[$offset+$col]." ",$width));
      }
      $t->put(BR=>1);

      $t->put(TEXT=>(' ' x $name_width));
      for (my $col=0; $col<$n_cols; $col++) {
	$t->put(TEXT=>&$trim($n[$offset+$col]." ",$width));
      }
      $t->put(BR=>1);

      $t->put(TEXT=>(' ' x $name_width));
      for (my $col=0; $col<$n_cols; $col++) {
	$t->put(TEXT=>&$trim("(".$gb->assignment_property($a[$offset+$col],"max").")",$width));
      }
      $t->put(BR=>1);

      foreach my $student (@student_keys) {
	my ($first,$last) = $gb->name($student);
	my $full_name = "$last, $first";
	if (length $full_name>$name_width-1) {$full_name=substr($full_name,0,$name_width-1)}
	$t->put(TEXT=>&$trim($full_name." ",$name_width));
        for (my $col=0; $col<$n_cols; $col++) {
	  my $j = $offset+$col;
	  my $grade = $gb->get_current_grade($student,$c[$j],$n[$j]);
	  if ($grade eq '') {$grade='--'}
	  $t->put(TEXT=>(sprintf "%${width}s",$grade." "));
        }
        $t->put(BR=>1);
      }

      $t->put(P=>1);

    }


    return ($t,$full_width);
}

sub student {
    my %args = (
		GB=>"",
		STUDENT=>"",
		FORMAT=>"plain",
                @_,
               );
    my $gb = $args{GB};
    my $format = $args{FORMAT};
    my $student = $args{STUDENT};
    my ($first,$last) = $gb->name($student);
    my $full_name = $first." ".$last;
    my $totals = Crunch::totals($gb,$student);
    my $t = Text->new($format);
    $t->put(P=>1,TEXT=>"$first $last, ".$gb->title());
    my @c = split(",",$gb->category_list());

    my $all = $totals->{"all"};
    $t->put(P=>1,TEXT=>"overall grade: ".format_percent_letter_and_fraction($gb,$all));
    foreach my $c(@c) {
      $t->put(BR=>1,TEXT=>$gb->category_name_plural($c).": ".format_percent_letter_and_fraction($gb,$totals->{$c}));
      my @a = split(",",$gb->assignment_list());
      my $sing = $gb->category_name_singular($c);
      if (!$gb->category_property_boolean($c,'single')) {
        foreach my $a(@a) {
  	  my ($cat,$ass) = (GradeBook::first_part_of_label($a),
			      GradeBook::second_part_of_label($a));
          if ($cat eq $c) {
	    my $grade = $gb->get_current_grade($student,$cat,$ass);
            if ($grade eq "") {$grade="--"}
	    my $max = $gb->assignment_property($a,"max");
	    my $ass_name = $gb->assignment_name($cat,$ass);
            $t->put(TEXT=>"$ass_name: $grade/$max",INDENTATION=>1,BR=>1);
          }
        } # end foreach assignment
      } # end if not single-assignment cat
    }
    return $t;
}

1;
